/*****************************************************************************
*
*	File		ANAM.java
*	Author		Vijay Sekhri
*	Date		Mon Sep 15 14:58:17 CDT 2003
*	Organization	Fermi National Accelerator Laboratory
*	Description	A class for authentication and authorization methods.
*	Email		sekhri@fnal.gov
*
*
******************************************************************************/
package gov.fnal.vox.saz.bl.util;


import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.net.Socket;
import java.net.InetAddress;
import java.security.cert.X509Certificate;

import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSException;
import org.ietf.jgss.Oid;

import org.gridforum.jgss.ExtendedGSSContext;
import org.gridforum.jgss.ExtendedGSSManager;

import org.globus.gsi.GSIConstants;
import org.globus.gsi.TrustedCertificates;
import org.globus.gsi.X509Credential;
import org.globus.gsi.CredentialException;
import org.globus.gsi.gssapi.SSLUtil;
import org.globus.gsi.gssapi.GSSConstants;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.gsi.gssapi.GlobusGSSManagerImpl;
import org.globus.gsi.gssapi.net.GssSocketFactory;
import org.globus.gsi.gssapi.net.impl.GSIGssSocket;
import org.globus.gsi.gssapi.net.GssSocket;
import org.globus.gsi.gssapi.auth.NoAuthorization;

import org.apache.log4j.Logger;
import gov.fnal.vox.saz.bl.util.ReadWriteSocket;

public class ANAM{
	protected String[] dn=null;
	protected String[] ca=null;
	protected int lifeTime=0;
	private String serverCert;
        private String serverKey;
	private String caCertDir=SAZConfiguration.getInstance().getDefaultCADir();
	//private TrustedCertificates trustedCerts;
	private TrustedCertificates trustedCerts= SAZConfiguration.getInstance().getTrustedCerts();
	private GSSCredential cred;
	private GSSManager manager;
	private static final transient Logger logger=Logger.getLogger(ANAM.class.getName());
	
	/**
	* Perform a gsi handshake and authenticate the two way connection. This is used at server side.
	* @param clientSocket The socket on which the handshake will be performed.
	* @param context The context which is initiated between the client and this server.
	* @return true - means the authentication is successful. false - means authentication failed.
	*/
	public boolean gsiAuthenticate(Socket clientSocket,ExtendedGSSContext context)  throws Exception{
		byte [] inToken=null;
		byte [] outToken=null;
		boolean toReturn=false;
		InputStream in=null;
		
		try{
			in = clientSocket.getInputStream();
			ReadWriteSocket rwSocket=new ReadWriteSocket(clientSocket);
	
			while (!context.isEstablished()){
				//inToken=SSLUtil.readSslMessage(in);
				inToken=rwSocket.readToken();
				outToken=context.acceptSecContext(inToken, 0, inToken.length);
				if (outToken != null){
					rwSocket.writeToken(outToken);
				}
			}
			logger.info("Context established.");
			//dn=context.getSrcName().toString();
			X509Certificate[] chain=(X509Certificate[])context.inquireByOid(GSSConstants.X509_CERT_CHAIN);
			//X509Certificate[] chain=(X509Certificate[])context.getCertificateChain();
			if(chain!=null){
				int chainLength=chain.length;
				dn=new String[chainLength];
				ca=new String[chainLength];
				for(int i=0;i<chainLength;i++){
					if(chain[i]!=null){
						dn[i]=chain[i].getSubjectDN().getName();
						ca[i]=chain[i].getIssuerDN().getName();
						logger.info("Initiator DN["+i+"]-->"+dn[i]);
						logger.info("Initiator CA["+i+"]-->"+ca[i]);
					}
				}
			}
			lifeTime=context.getLifetime();
			logger.info("Acceptor  : " + context.getTargName());
			logger.info("Lifetime  : " + lifeTime);
			logger.info("Privacy   : " + context.getConfState());
			GlobusGSSCredentialImpl cred =(GlobusGSSCredentialImpl)context.getDelegCred();
			logger.info("Delegated credential :");
			if (cred != null){
				logger.info("Not null");
			}else{
				logger.info("None");
			}
			toReturn=true;
		}finally{
			try{
				context.dispose();
			}catch(Exception e1){
				e1.printStackTrace();
			}
		}		

		return(toReturn);

	}


	/**
	* User interface for defining the authorize method<br>
	*/
	public boolean authorize(){
		return(true);
	}
	

	/**
	* Sends a request to the server to perform gsi authentication and autherization (if enabled on server) on the given socket. This method is used at client side to initiate handshake.
	* @param clientSocket the socket on which the mutual authentication will be performed
	* @param context The context which is initiated between this client and the server.
	* @return true - means the auhentication (and autherization) is successful. false - means authentication (and/or autherization) failed.
	*/
	public boolean sendGsiRequest(Socket clientSocket,ExtendedGSSContext context) throws Exception{
		InputStream in=null;
		byte [] inToken=null;
		byte [] outToken=null;
		try{
			in = clientSocket.getInputStream();
			ReadWriteSocket rwSocket=new ReadWriteSocket(clientSocket);
	    		inToken = new byte[0];
			outToken = null;
			while (!context.isEstablished()){
				outToken=context.initSecContext(inToken, 0, inToken.length);
				if (outToken != null) {
					rwSocket.writeToken(outToken);
				}
				if (!context.isEstablished()) {
					//inToken = SSLUtil.readSslMessage(in);
					inToken=rwSocket.readToken();
				}
			}
			return(true);
		}finally{
			try{
				context.dispose();
			}catch(Exception e1){
				e1.printStackTrace();
			}
		}	
	}

	protected ExtendedGSSContext createGSSContext() throws GSSException{
		ExtendedGSSContext context= (ExtendedGSSContext)manager.createContext(cred);
		logger.debug("After creating GSSContext");
		context.setOption(GSSConstants.GSS_MODE,GSIConstants.MODE_GSI);
		logger.debug("After setting MODE_GSI");
		if (trustedCerts != null) {
			context.setOption(GSSConstants.TRUSTED_CERTIFICATES,trustedCerts);
			logger.debug("After setting TrustedCerts");
		}
		logger.debug("Returning Context");
		return context;
	}

	/**
        * A method to instantiate a ExtendedGSSContext object This is used at the server side to before calling gsiAuthenticate
	* @param cert The host certificate to key. If NULL then by default it uses /etc/grid-security/hostcert.pem
	* @param key The host key to user. If NULL then by default it uses /etc/grid-security/hostkey.pem
	* @return An ExtendedGSSContext object that can used for handshake , wraping and unwrapping messages. 
	*/
	public ExtendedGSSContext getGsiServerContext(String cert,String key) throws IOException{
		X509Credential serverCred=null;
		ExtendedGSSContext context=null;
		try{
			logger.info("cert: "+cert+" key: "+key);
			serverCred =new X509Credential(cert,key);
			if(serverCred != null){
				logger.debug("Server credentials are not null");
				//serverCred.verify();
			}else{
				logger.debug("server credentials are null");
			}
			cred = new GlobusGSSCredentialImpl(serverCred, GSSCredential.INITIATE_AND_ACCEPT);
			logger.debug("After creating GSS Credential..");
			if (caCertDir != null){
				logger.info("caCertDir: "+caCertDir);
				//trustedCerts=TrustedCertificates.load(caCertDir);
			}
			manager = ExtendedGSSManager.getInstance();
			context=createGSSContext();
        		return(context);
		}catch(CredentialException ce){
                        logger.error(ce.getMessage());
                        throw new IOException("Could not load host globus credentials "+ce.toString());
                }catch(GSSException e){
                        throw new IOException("Could not load host globus credentials "+e.toString());
                }catch(Exception e){
			logger.info(e.toString());
		}finally{
			try{
				logger.debug("Before disposing context..");
				context.dispose();
				logger.debug("After disposing context..");
			}catch(Exception e1){
				logger.error(e1.getMessage());
			}
		}
		return(context);
	}

	public ExtendedGSSContext getGsiClientContext() throws GSSException{
		GSSManager manager = new GlobusGSSManagerImpl();
		ExtendedGSSContext context = (ExtendedGSSContext)manager.createContext(null,GSSConstants.MECH_OID,null,GSSContext.DEFAULT_LIFETIME);
		context.setOption(GSSConstants.GSS_MODE,GSIConstants.MODE_GSI);
		//context.requestCredDeleg(true);
		context.requestCredDeleg(false);
		return(context);

	}

	public Socket getGsiServerSocket(Socket clientSocket,ExtendedGSSContext context) throws Exception{
		//InputStream in=null;
		//OutputStream out=null;
		logger.debug("Entered getGsiServerSocket..");
		clientSocket = GssSocketFactory.getDefault().createSocket(clientSocket, null, 0, context);
		logger.debug("After creating client socket..");
		((GssSocket)clientSocket).setUseClientMode(false);
		logger.debug("After setting useClientMode to false..");
		((GssSocket)clientSocket).setWrapMode(GssSocket.GSI_MODE);
		logger.debug("After setting wrap mode ..");
		((GssSocket)clientSocket).setAuthorization(NoAuthorization.getInstance());
		logger.debug("After setting authorization ..");
		/*logger.debug("ANAM - Before getting OutputStream");
                out=clientSocket.getOutputStream();
                logger.debug("ANAM - After getting OutputStream");
		logger.debug("ANAM - Before getting InputStream");
                in=clientSocket.getInputStream();
		logger.debug("ANAM - After getting InputStream");*/
		return(clientSocket);
	}

	public Socket getGsiClientSocket(String host,int port,ExtendedGSSContext context) throws Exception{
		Socket clientSocket = GssSocketFactory.getDefault().createSocket(host, port, context);
		((GssSocket)clientSocket).setWrapMode(GssSocket.GSI_MODE);
		((GssSocket)clientSocket).setAuthorization(NoAuthorization.getInstance());
		return(clientSocket);
	}

}
